@@ -19,8 +19,9 @@ from concurrent.futures import ProcessPoolExecutor, as_completed
 from copy import copy
 from functools import partial
 from itertools import repeat, product, chain
+from math import copysign
 from numbers import Number
-from typing import Callable, Union, List, Optional, Sequence, Tuple, Type
+from typing import Callable, Dict, List, Optional, Sequence, Tuple, Type, Union
 
 import numpy as np
 import pandas as pd
@@ -198,7 +199,7 @@ class Strategy(metaclass=ABCMeta):
             sl: float = None,
             tp: float = None):
         """
-        Place a new short order. For explanation of parameters, see `Order` and its properties.
+        Place a new long order. For explanation of parameters, see `Order` and its properties.
 
         See also `Strategy.sell()`.
         """
@@ -222,7 +223,7 @@ class Strategy(metaclass=ABCMeta):
         return self._broker.new_order(-size, limit, stop, sl, tp)
 
     @property
-    def equity(self):
+    def equity(self) -> float:
         """Current account equity (cash plus assets)."""
         return self._broker.equity
 
@@ -261,17 +262,17 @@ class Strategy(metaclass=ABCMeta):
         return self._broker.position
 
     @property
-    def orders(self) -> '_Orders[Order]':
+    def orders(self) -> 'Tuple[Order, ...]':
         """List of orders (see `Order`) waiting for execution."""
         return _Orders(self._broker.orders)
 
     @property
-    def trades(self) -> 'Tuple[Trade]':
+    def trades(self) -> 'Tuple[Trade, ...]':
         """List of active trades (see `Trade`)."""
         return tuple(self._broker.trades)
 
     @property
-    def closed_trades(self) -> 'Tuple[Trade]':
+    def closed_trades(self) -> 'Tuple[Trade, ...]':
         """List of settled trades (see `Trade`)."""
         return tuple(self._broker.closed_trades)
 
@@ -314,17 +315,17 @@ class Position:
         return self.size != 0
 
     @property
-    def size(self):
+    def size(self) -> float:
         """Position size in units of asset. Negative if position is short."""
         return sum(trade.size for trade in self.__broker.trades)
 
     @property
-    def pl(self):
+    def pl(self) -> float:
         """Profit (positive) or loss (negative) of the current position in cash units."""
         return sum(trade.pl for trade in self.__broker.trades)
 
     @property
-    def pl_pct(self):
+    def pl_pct(self) -> float:
         """Profit (positive) or loss (negative) of the current position in percent."""
         weights = np.abs([trade.size for trade in self.__broker.trades])
         weights = weights / weights.sum()
@@ -332,12 +333,12 @@ class Position:
         return (pl_pcts * weights).sum()
 
     @property
-    def is_long(self):
+    def is_long(self) -> bool:
         """True if the position is long (position size is positive)."""
         return self.size > 0
 
     @property
-    def is_short(self):
+    def is_short(self) -> bool:
         """True if the position is short (position size is negative)."""
         return self.size < 0
 
@@ -379,6 +380,7 @@ class Order:
                  tp_price: float = None,
                  parent_trade: 'Trade' = None):
         self.__broker = broker
+        assert size != 0
         self.__size = size
         self.__limit_price = limit_price
         self.__stop_price = stop_price
@@ -417,7 +419,7 @@ class Order:
     # Fields getters
 
     @property
-    def size(self):
+    def size(self) -> float:
         """
         Order size (negative for short orders).
 
@@ -428,7 +430,7 @@ class Order:
         return self.__size
 
     @property
-    def limit(self) -> float:
+    def limit(self) -> Optional[float]:
         """
         Order limit price for [limit orders], or None for [market orders],
         which are filled at next available price.
@@ -439,7 +441,7 @@ class Order:
         return self.__limit_price
 
     @property
-    def stop(self):
+    def stop(self) -> Optional[float]:
         """
         Order stop price for [stop-limit/stop-market][_] order,
         otherwise None if no stop was set, or the stop price has already been hit.
@@ -449,7 +451,7 @@ class Order:
         return self.__stop_price
 
     @property
-    def sl(self):
+    def sl(self) -> Optional[float]:
         """
         A stop-loss price at which, if set, a new contingent stop-market order
         will be placed upon the `Trade` following this order's execution.
@@ -458,7 +460,7 @@ class Order:
         return self.__sl_price
 
     @property
-    def tp(self):
+    def tp(self) -> Optional[float]:
         """
         A take-profit price at which, if set, a new contingent limit order
         will be placed upon the `Trade` following this order's execution.
@@ -508,9 +510,9 @@ class Trade:
         self.__broker = broker
         self.__size = size
         self.__entry_price = entry_price
-        self.__entry_bar = entry_bar
-        self.__exit_price = None
-        self.__exit_bar = None
+        self.__exit_price = None  # type: Optional[float]
+        self.__entry_bar = entry_bar  # type: int
+        self.__exit_bar = None  # type: Optional[int]
         self.__sl_order = None  # type: Optional[Order]
         self.__tp_order = None  # type: Optional[Order]
 
@@ -530,9 +532,9 @@ class Trade:
     def close(self, portion: float = 1.):
         """Place new `Order` to close `portion` of the trade at next market price."""
         assert 0 < portion <= 1, "portion must be a fraction between 0 and 1"
-        # TODO: check this insert
-        self.__broker.orders.insert(0, Order(self.__broker, -round(self.size * portion),
-                                             parent_trade=self))
+        size = copysign(max(1, round(abs(self.__size) * portion)), -self.__size)
+        order = Order(self.__broker, size, parent_trade=self)
+        self.__broker.orders.insert(0, order)
 
     # Fields getters
 
@@ -547,7 +549,7 @@ class Trade:
         return self.__entry_price
 
     @property
-    def exit_price(self) -> float:
+    def exit_price(self) -> Optional[float]:
         """Trade exit price (or None if the trade is still active)."""
         return self.__exit_price
 
@@ -557,7 +559,7 @@ class Trade:
         return self.__entry_bar
 
     @property
-    def exit_bar(self) -> int:
+    def exit_bar(self) -> Optional[int]:
         """
         Candlestick bar index of when the trade was exited
         (or None if the trade is still active).
@@ -574,10 +576,22 @@ class Trade:
 
     # Extra properties
 
+    @property
+    def entry_time(self) -> Union[pd.Timestamp, int]:
+        """Datetime of when the trade was entered."""
+        return self.__broker._data.index[self.__entry_bar]
+
+    @property
+    def exit_time(self) -> Optional[Union[pd.Timestamp, int]]:
+        """Datetime of when the trade was exited."""
+        if self.__exit_bar is None:
+            return None
+        return self.__broker._data.index[self.__exit_bar]
+
     @property
     def is_long(self):
         """True if the trade is long (trade size is positive)."""
-        return self.size > 0
+        return self.__size > 0
 
     @property
     def is_short(self):
@@ -594,7 +608,7 @@ class Trade:
     def pl_pct(self):
         """Trade profit (positive) or loss (negative) in percent."""
         price = self.__exit_price or self.__broker.last_price
-        return np.sign(self.__size) * (price / self.__entry_price - 1)
+        return copysign(1, self.__size) * (price / self.__entry_price - 1)
 
     @property
     def value(self):
@@ -615,6 +629,10 @@ class Trade:
         """
         return self.__sl_order and self.__sl_order.stop
 
+    @sl.setter
+    def sl(self, price: float):
+        self.__set_contingent('sl', price)
+
     @property
     def tp(self):
         """
@@ -626,12 +644,8 @@ class Trade:
         """
         return self.__tp_order and self.__tp_order.limit
 
-    @sl.setter
-    def sl(self, price):
-        self.__set_contingent('sl', price)
-
     @tp.setter
-    def tp(self, price):
+    def tp(self, price: float):
         self.__set_contingent('tp', price)
 
     def __set_contingent(self, type, price):
@@ -672,7 +686,7 @@ class _Broker:
             self._cash, self.position.pl, len(self.trades))
 
     def new_order(self,
-                  size: float = None,
+                  size: float,
                   limit: float = None,
                   stop: float = None,
                   sl: float = None,
@@ -691,14 +705,13 @@ class _Broker:
         is_long = size > 0
 
         if is_long:
-            # TODO: check stop as well
-            assert (sl or -np.inf) <= (limit or self.last_price) <= (tp or np.inf), \
-                "Long orders require: SL ({}) < LIMIT ({}) < TP ({})".format(
-                    sl, limit or self.last_price, tp)
+            if not (sl or -np.inf) <= (limit or stop or self.last_price) <= (tp or np.inf):
+                raise ValueError("Long orders require: SL ({}) < LIMIT ({}) < TP ({})".format(
+                    sl, limit or stop or self.last_price, tp))
         else:
-            assert (tp or -np.inf) <= (limit or self.last_price) <= (sl or np.inf), \
-                "Short orders require: TP ({}) < LIMIT ({}) < SL ({})".format(
-                    tp, limit or self.last_price, sl)
+            if not (tp or -np.inf) <= (limit or stop or self.last_price) <= (sl or np.inf):
+                raise ValueError("Short orders require: TP ({}) < LIMIT ({}) < SL ({})".format(
+                    tp, limit or stop or self.last_price, sl))
 
         order = Order(self, size, limit, stop, sl, tp, trade)
         # Put the new order in the order queue,
@@ -720,18 +733,18 @@ class _Broker:
         return order
 
     @property
-    def last_price(self):
+    def last_price(self) -> float:
         """Return price at the last (current) close.
         Used e.g. in `Orders._is_price_ok()` to see if the set price is reasonable.
         """
         return self._data.Close[-1]
 
     @property
-    def equity(self):
+    def equity(self) -> float:
         return self._cash + sum(trade.pl for trade in self.trades)
 
     @property
-    def margin_available(self):
+    def margin_available(self) -> float:
         # From https://github.com/QuantConnect/Lean/pull/3768
         margin_used = sum(trade.value / self._leverage for trade in self.trades)
         return max(0, self.equity - margin_used)
@@ -757,6 +770,7 @@ class _Broker:
         data = self._data
         open, high, low = data.Open[-1], data.High[-1], data.Low[-1]
         prev_close = data.Close[-2]
+        reprocess_orders = False
 
         # Process orders
         for order in list(self.orders):  # type: Order
@@ -807,17 +821,18 @@ class _Broker:
             # If order is a SL/TP order, it should close an existing trade it was contingent upon
             if order.parent_trade:
                 trade = order.parent_trade
+                _prev_size = trade.size
                 # If this trade isn't already closed (e.g. on multiple `trade.close(.5)` calls)
                 if trade in self.trades:
                     self._reduce_trade(trade, price, order.size, time_index)
-                    assert order.size != -trade.size or trade not in self.trades
+                    assert order.size != -_prev_size or trade not in self.trades
                 if order in (trade._sl_order,
                              trade._tp_order):
                     assert order.size == -trade.size
                     assert order not in self.orders  # Removed when trade was closed
                 else:
                     # It's a trade.close() order, now done
-                    assert abs(trade.size) >= abs(order.size) >= 1
+                    assert abs(_prev_size) >= abs(order.size) >= 1
                     self.orders.remove(order)
                 continue
 
@@ -825,14 +840,14 @@ class _Broker:
 
             # Adjust price to include commission (or bid-ask spread).
             # In long positions, the adjusted price is a fraction higher, and vice versa.
-            adjusted_price = price * (1 + (np.sign(order.size) * self._commission))
+            adjusted_price = price * (1 + copysign(self._commission, order.size))
 
             # If order size was specified proportionally,
             # precompute true size in units, accounting for margin and spread/commissions
             size = order.size
             if -1 < size < 1:
-                size = np.sign(size) * int((self.margin_available * self._leverage * abs(size))
-                                           // adjusted_price)
+                size = copysign(int((self.margin_available * self._leverage * abs(size))
+                                    // adjusted_price), size)
                 # Not enough cash/margin even for a single unit
                 if not size:
                     self.orders.remove(order)
@@ -847,7 +862,7 @@ class _Broker:
                 for trade in list(self.trades):
                     if trade.is_long == order.is_long:
                         continue
-                    assert np.sign(trade.size) + np.sign(order.size) == 0
+                    assert trade.size * order.size < 0
 
                     # Order size greater than this opposite-directed existing trade,
                     # so it will be closed completely
@@ -870,11 +885,31 @@ class _Broker:
             if need_size:
                 self._open_trade(adjusted_price, need_size, order.sl, order.tp, time_index)
 
+                # We need to reprocess the SL/TP orders newly added to the queue.
+                # This allows e.g. SL hitting in the same bar the order was open.
+                # See https://github.com/kernc/backtesting.py/issues/119
+                if order.sl or order.tp:
+                    if is_market_order:
+                        reprocess_orders = True
+                    elif (low <= (order.sl or -np.inf) <= high or
+                          low <= (order.tp or -np.inf) <= high):
+                        warnings.warn(
+                            "A SL/TP order would execute in the same bar as its contingent upon "
+                            "stop/limit order. Since we can't assert the precise intra-candle "
+                            "price movement, the affected SL/TP order will be executed on "
+                            "the next (matching) price/bar, making the result (of this trade) "
+                            "somewhat dubious. "
+                            "See https://github.com/kernc/backtesting.py/issues/119",
+                            UserWarning)
+
             # Order processed
             self.orders.remove(order)
 
+        if reprocess_orders:
+            self._process_orders()
+
     def _reduce_trade(self, trade: Trade, price: float, size: float, time_index: int):
-        assert np.sign(trade.size) != np.sign(size)
+        assert trade.size * size < 0
         assert abs(trade.size) >= abs(size)
 
         size_left = trade.size + size
@@ -948,7 +983,9 @@ class Backtest:
 
             df['Open'] = df['High'] = df['Low'] = df['Close']
 
-        DataFrame index can be either datetime index (timestamps)
+        The passed data frame can contain additional columns that
+        can be used by the strategy (e.g. sentiment info).
+        DataFrame index can be either a datetime index (timestamps)
         or a monotonic range index (i.e. a sequence of periods).
 
         `strategy` is a `backtesting.backtesting.Strategy`
@@ -956,9 +993,9 @@ class Backtest:
 
         `cash` is the initial cash to start with.
 
-        `commission` is the commision ratio. E.g. if your broker's commission
+        `commission` is the commission ratio. E.g. if your broker's commission
         is 1% of trade value, set commission to `0.01`. Note, if you wish to
-        account for bid-ask spread, you cam approximate doing so by increasing
+        account for bid-ask spread, you can approximate doing so by increasing
         the commission, e.g. set it to `0.0002` for commission-less forex
         trading where the average spread is roughly 0.2‰ of asking price.
 
@@ -975,6 +1012,10 @@ class Backtest:
         If `False`, the opposite-facing orders first close existing trades in
         a [FIFO] manner.
 
+        If `exclusive_orders` is `True`, each new order auto-closes the previous
+        trade/position, making at most a single trade (long or short) in effect
+        at each time.
+
         [FIFO]: https://www.investopedia.com/terms/n/nfa-compliance-rule-2-43b.asp
         """
 
@@ -1155,8 +1196,9 @@ class Backtest:
         if not kwargs:
             raise ValueError('Need some strategy parameters to optimize')
 
+        maximize_key = None
         if isinstance(maximize, str):
-
+            maximize_key = str(maximize)
             stats = self._results if self._results is not None else self.run()
             if maximize not in stats:
                 raise ValueError('`maximize`, if str, must match a key in pd.Series '
@@ -1200,6 +1242,7 @@ class Backtest:
                           stacklevel=2)
 
         heatmap = pd.Series(np.nan,
+                            name=maximize_key,
                             index=pd.MultiIndex.from_tuples([p.values() for p in param_combos],
                                                             names=next(iter(param_combos)).keys()))
 
@@ -1219,7 +1262,7 @@ class Backtest:
         # in a copy-on-write manner, achieving better performance/RAM benefit.
         backtest_uuid = np.random.random()
         param_batches = list(_batch(param_combos))
-        Backtest._mp_backtests[backtest_uuid] = (self, param_batches, maximize)
+        Backtest._mp_backtests[backtest_uuid] = (self, param_batches, maximize)  # type: ignore
         try:
             # If multiprocessing start method is 'fork' (i.e. on POSIX), use
             # a pool of processes to compute results in parallel.
@@ -1248,7 +1291,7 @@ class Backtest:
         if pd.isnull(best_params):
             # No trade was made in any of the runs. Just make a random
             # run so we get some, if empty, results
-            self.run(**param_combos[0])
+            self.run(**param_combos[0])  # type: ignore
         else:
             # Re-run best strategy so that the next .plot() call will render it
             self.run(**dict(zip(heatmap.index.names, best_params)))
@@ -1264,7 +1307,7 @@ class Backtest:
                              for stats in (bt.run(**params)
                                            for params in param_batches[batch_index])]
 
-    _mp_backtests = {}
+    _mp_backtests = {}  # type: Dict[float, Tuple[Backtest, List, Callable]]
 
     @staticmethod
     def _compute_drawdown_duration_peaks(dd: pd.Series):
@@ -1303,9 +1346,9 @@ class Backtest:
             'ExitPrice': [t.exit_price for t in trades],
             'PnL': [t.pl for t in trades],
             'ReturnPct': [t.pl_pct for t in trades],
+            'EntryTime': [t.entry_time for t in trades],
+            'ExitTime': [t.exit_time for t in trades],
         })
-        trades_df['EntryTime'] = trades_df['EntryBar'].apply(index.__getitem__)
-        trades_df['ExitTime'] = trades_df['ExitBar'].apply(index.__getitem__)
         trades_df['Duration'] = trades_df['ExitTime'] - trades_df['EntryTime']
 
         pl = trades_df['PnL']
@@ -1325,7 +1368,7 @@ class Backtest:
 
         have_position = np.repeat(0, len(index))
         for t in trades:
-            have_position[t.entry_bar:t.exit_bar + 1] = 1
+            have_position[t.entry_bar:t.exit_bar + 1] = 1  # type: ignore
 
         s.loc['Exposure Time [%]'] = have_position.mean() * 100  # In "n bars" time, not index time
         s.loc['Equity Final [$]'] = equity[-1]
