@@ -3,4 +3,4 @@ import unittest
 suite = unittest.defaultTestLoader.discover('backtesting.test',
                                             pattern='_test*.py')
 if __name__ == '__main__':
-    unittest.TextTestRunner().run(suite)
+    unittest.TextTestRunner(verbosity=2).run(suite)
