@@ -1,3 +1,4 @@
+import inspect
 import os
 import sys
 import time
@@ -192,7 +193,9 @@ class TestBacktest(TestCase):
                         assert trade.is_short
                         assert trade.size < 0
                         assert trade.entry_bar > 0
+                        assert isinstance(trade.entry_time, pd.Timestamp)
                         assert trade.exit_bar is None
+                        assert trade.exit_time is None
                         assert trade.entry_price > 0
                         assert trade.exit_price is None
                         assert trade.pl / 1
@@ -208,7 +211,8 @@ class TestBacktest(TestCase):
                         self.position.close()
 
         bt = Backtest(GOOG, Assertive)
-        stats = bt.run()
+        with self.assertWarns(UserWarning):
+            stats = bt.run()
         self.assertEqual(stats['# Trades'], 144)
 
     def test_broker_params(self):
@@ -327,6 +331,52 @@ class TestBacktest(TestCase):
                 self.assertFalse(stats['_equity_curve']['Equity'].isnull().any())
                 self.assertEqual(stats['_strategy'].__class__, strategy)
 
+    def test_trade_enter_hit_sl_on_same_day(self):
+        the_day = pd.Timestamp("2012-10-17 00:00:00")
+
+        class S(Strategy):
+            def init(self): pass
+
+            def next(self):
+                if self.data.index[-1] == the_day:
+                    self.buy(sl=720)
+
+        self.assertEqual(Backtest(GOOG, S).run()._trades.iloc[0].ExitPrice, 720)
+
+        class S(S):
+            def next(self):
+                if self.data.index[-1] == the_day:
+                    self.buy(stop=758, sl=720)
+
+        with self.assertWarns(UserWarning):
+            self.assertEqual(Backtest(GOOG, S).run()._trades.iloc[0].ExitPrice, 705.58)
+
+    def test_stop_price_between_sl_tp(self):
+        class S(Strategy):
+            def init(self): pass
+
+            def next(self):
+                if self.data.index[-1] == pd.Timestamp("2004-09-09 00:00:00"):
+                    self.buy(stop=104, sl=103, tp=110)
+
+        with self.assertWarns(UserWarning):
+            self.assertEqual(Backtest(GOOG, S).run()._trades.iloc[0].EntryPrice, 104)
+
+    def test_position_close_portion(self):
+        class SmaCross(Strategy):
+            def init(self):
+                self.sma1 = self.I(SMA, self.data.Close, 10)
+                self.sma2 = self.I(SMA, self.data.Close, 20)
+
+            def next(self):
+                if not self.position and crossover(self.sma1, self.sma2):
+                    self.buy(size=10)
+                if self.position and crossover(self.sma2, self.sma1):
+                    self.position.close(portion=.5)
+
+        bt = Backtest(GOOG, SmaCross, commission=.002)
+        bt.run()
+
 
 class TestStrategy(TestCase):
     def _Backtest(self, strategy_coroutine, **kwargs):
@@ -400,7 +450,8 @@ class TestOptimize(TestCase):
         res = bt.optimize(**OPT_PARAMS)
         self.assertIsInstance(res, pd.Series)
 
-        res2 = bt.optimize(**OPT_PARAMS, maximize=lambda s: s['SQN'])
+        default_maximize = inspect.signature(Backtest.optimize).parameters['maximize'].default
+        res2 = bt.optimize(**OPT_PARAMS, maximize=lambda s: s[default_maximize])
         self.assertDictEqual(res.filter(regex='^[^_]').fillna(-1).to_dict(),
                              res2.filter(regex='^[^_]').fillna(-1).to_dict())
 
@@ -408,6 +459,7 @@ class TestOptimize(TestCase):
                                     constraint=lambda d: d.slow > 2 * d.fast)
         self.assertIsInstance(heatmap, pd.Series)
         self.assertEqual(len(heatmap), 4)
+        self.assertEqual(heatmap.name, default_maximize)
 
         with _tempfile() as f:
             bt.plot(filename=f, open_browser=False)
@@ -696,13 +748,15 @@ class TestUtil(TestCase):
             pass
 
         class Class:
-            pass
+            def __call__(self):
+                pass
 
         self.assertEqual(_as_str('4'), '4')
         self.assertEqual(_as_str(4), '4')
         self.assertEqual(_as_str(_Indicator([1, 2], name='x')), 'x')
         self.assertEqual(_as_str(func), 'func')
         self.assertEqual(_as_str(Class), 'Class')
+        self.assertEqual(_as_str(Class()), 'Class')
         self.assertEqual(_as_str(pd.Series([1, 2], name='x')), 'x')
         self.assertEqual(_as_str(pd.DataFrame()), 'df')
         self.assertEqual(_as_str(lambda x: x), 'λ')
